#![deny(unsafe_code)]
#![warn(clippy::pedantic, clippy::cargo)]

use proc_macro::TokenStream;
use quote::quote;
use syn::{FnArg, ItemFn};

#[proc_macro_attribute]
pub fn test(_args: TokenStream, tokens: TokenStream) -> TokenStream {
    let ItemFn {
        attrs,
        vis,
        mut sig,
        block,
    } = match syn::parse(tokens) {
        Ok(inner) => inner,
        Err(e) => {
            return TokenStream::from(
                syn::Error::new(e.span(), "teto::test can only be called on functions")
                    .to_compile_error(),
            )
        }
    };

    let inputs = std::mem::take(&mut sig.inputs);
    // Since this macro replaces #[test], we can assume that the only arg (if
    // any) is the token arg.
    let test_token_name = inputs.first().and_then(|fn_arg| match fn_arg {
        FnArg::Receiver(_) => None,
        FnArg::Typed(typed) => Some(&typed.ty),
    });
    let inner_fn_invocation = if let Some(test_token_name) = test_token_name {
        quote! {
            let token = const { #test_token_name::__private_teto_macros_only__create_teto_test_token };
            inner(token())
        }
    } else {
        quote! {
            inner()
        }
    };

    // Maybe I need to do this for asyncness and unsafety?
    let maybe_const = &sig.constness;

    let wrapped = quote! {
        #(#attrs)*
        #[test]
        #vis #sig {
            #maybe_const fn inner(#inputs)
                #block

            #inner_fn_invocation
        }

    };

    TokenStream::from(wrapped)
}

// #[proc_macro_attribute]
// pub fn wrapped(_args: TokenStream, tokens: TokenStream) -> TokenStream {
//     todo!()
// }
